/* =============================================================
 * File  : SRTtypeteletext.cpp
 * Author: Peter Holmes (pholmes@zetron.com)
 *         Taken from the original OSDtypeteletext.cpp by
 *              Frank Muenchow <beebof@gmx.de>
 *              Martin Barnasconi <martin@barnasconi.net>
 * Date  : 2007-11-24
 *
 * This program is free software; you can redistribute it
 * and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation;
 * either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * =============================================================
 */

#include <cmath>

#include <qmutex.h>

#include "srttypeteletext.h"
#include "mpeg2fix.h"
#include "../libs/libmythtv/vbilut.h"

const int    SRTTypeTeletext::kTeletextColumns = 40;
const int    SRTTypeTeletext::kTeletextRows    = 26;

/*****************************************************************************/
SRTTypeTeletext::SRTTypeTeletext()
    : 
      // last fetched page
      m_fetchpage(0),                   m_fetchsubpage(0),
      // currently displayed page
      m_curpage(0x100),                 m_cursubpage(-1),
      m_curpage_showheader(true),       m_curpage_issubtitle(false),

      m_transparent(false),             m_revealHidden(false),
      m_displaying(false),              m_changed(false),
      m_header_changed(false),          m_page_changed(false),
      m_output_device(NULL)
{
    // fill Bitswap
    for (int i = 0; i < 256; i++)
    {
        m_bitswap[i] = 0;
        for (int bit = 0; bit < 8; bit++)
            if (i & (1 << bit))
                m_bitswap[i] |= (1 << (7-bit));
    }

    Reset(); // initializes m_magazines
}

/** \fn SRTTypeTeletext::Reset(void)
 *  \brief Resets the teletext data cache
 */
void SRTTypeTeletext::Reset(void)
{
    for (uint mag = 0; mag < 8; mag++)
    {
        // clear all sub pages in page
        int_to_page_t::iterator iter;
        iter = m_magazines[mag].pages.begin();
        while (iter != m_magazines[mag].pages.end())
        {
            TeletextPage *page = &iter->second;
            page->subpages.clear();
            ++iter;
        }

        // clear pages
        m_magazines[mag].pages.clear();
        m_magazines[mag].current_page = 0;
        m_magazines[mag].current_subpage = 0;
        m_magazines[mag].loadingpage.active = false;
    }
    memset(m_header, ' ', 40);

    m_curpage    = 0x100;
    m_cursubpage = -1;
    m_curpage_showheader = true;

    m_pageinput[0] = '1';
    m_pageinput[1] = '0';
    m_pageinput[2] = '0';
}

/** \fn SRTTypeTeletext::AddPageHeader(int,int,const unsigned char*,int,int,int)
 *  \brief Adds a new page header (page=0 if none).
 */
void SRTTypeTeletext::AddPageHeader(int page, int subpage,
                                    const unsigned char* buf,
                                    int vbimode, int lang, int flags)
{
    int magazine = MAGAZINE(page);
    if (magazine < 1 || magazine > 8)
        return;
    int lastPage = m_magazines[magazine - 1].current_page;
    int lastSubPage = m_magazines[magazine - 1].current_subpage;

    // update the last fetched page if the magazine is the same
    // and the page no. is different

    if ((page != lastPage || subpage != lastSubPage) &&
        m_magazines[magazine - 1].loadingpage.active)
    {
        TeletextSubPage *ttpage = FindSubPage(lastPage, lastSubPage);
        if (!ttpage)
        {
            ttpage = &(m_magazines[magazine - 1]
                       .pages[lastPage].subpages[lastSubPage]);
            m_magazines[magazine - 1].pages[lastPage].pagenum = lastPage;
            ttpage->subpagenum = lastSubPage;
        }

        memcpy(ttpage, &m_magazines[magazine - 1].loadingpage,
               sizeof(TeletextSubPage));

        m_magazines[magazine - 1].loadingpage.active = false;

        PageUpdated(lastPage, lastSubPage);
    }

    m_fetchpage = page;
    m_fetchsubpage = subpage;

    TeletextSubPage *ttpage = &m_magazines[magazine - 1].loadingpage;

    m_magazines[magazine - 1].current_page = page;
    m_magazines[magazine - 1].current_subpage = subpage;

    memset(ttpage->data, ' ', sizeof(ttpage->data));

    ttpage->active = true;
    ttpage->subpagenum = subpage;

    for (uint i = 0; i < 6; i++)
        ttpage->floflink[i] = 0;

    ttpage->lang = lang;
    ttpage->flags = flags;
    ttpage->flof = 0;

    ttpage->subtitle = (vbimode == VBI_DVB_SUBTITLE);

    memset(ttpage->data[0], ' ', 8 * sizeof(uint8_t));

    if (vbimode == VBI_DVB || vbimode == VBI_DVB_SUBTITLE)
    {
        for (uint j = 8; j < 40; j++)
            ttpage->data[0][j] = m_bitswap[buf[j]];
    }
    else 
    {
        memcpy(ttpage->data[0]+0, buf, 40);
    }
    
    if ( !(ttpage->flags & TP_INTERRUPTED_SEQ)) 
    {
        memcpy(m_header, ttpage->data[0], 40);
        HeaderUpdated(ttpage->data[0],ttpage->lang);
    }
}

/** \fn SRTTypeTeletext::AddTeletextData(int,int,const unsigned char*,int)
 *  \brief Adds Teletext Data from TeletextDecoder
 */
void SRTTypeTeletext::AddTeletextData(int magazine, int row, 
                                      const unsigned char* buf, int vbimode)
{
    int b1, b2, b3, err;

    if (magazine < 1 || magazine > 8)
        return;

    int currentpage = m_magazines[magazine - 1].current_page;
    if (!currentpage)
        return;

    TeletextSubPage *ttpage = &m_magazines[magazine - 1].loadingpage;

    switch (row)
    {
        case 1 ... 24: // Page Data
            if (vbimode == VBI_DVB || vbimode == VBI_DVB_SUBTITLE)
            {
                for (uint j = 0; j < 40; j++)
                    ttpage->data[row][j] = m_bitswap[buf[j]];
            } 
            else 
            {
                memcpy(ttpage->data[row], buf, 40);
            }
            break;
        case 26:
            /* XXX TODO: Level 1.5, 2.5, 3.5
            *      Character location & override
            * Level 2.5, 3.5
            *      Modifying display attributes
            * All levels
            *      VCR Programming
            * See 12.3
            */
            break;
        case 27: // FLOF data (FastText)
            switch (vbimode)
            {
                case VBI_IVTV:
                    b1 = hamm8(buf, &err);
                    b2 = hamm8(buf + 37, &err);
                    if (err & 0xF000)
                        return;
                     break;
                case VBI_DVB:
                case VBI_DVB_SUBTITLE:
                    b1 = hamm84(buf, &err);
                    b2 = hamm84(buf + 37, &err);
                    if (err == 1)
                        return;
                    break;
                default:
                    return;
            }
            if (b1 != 0 || not(b2 & 8))
                return;

            for (int i = 0; i < 6; ++i)
            {
                err = 0;
                switch (vbimode)
                {
                    case VBI_IVTV:
                        b1 = hamm16(buf+1+6*i, &err);
                        b2 = hamm16(buf+3+6*i, &err);
                        b3 = hamm16(buf+5+6*i, &err);
                        if (err & 0xF000)
                            return;
                        break;
                    case VBI_DVB:
                    case VBI_DVB_SUBTITLE:
                        b1 = hamm84(buf+2+6*i, &err) * 16 +
                        hamm84(buf+1+6*i, &err);
                        b2 = hamm84(buf+4+6*i, &err) * 16 +
                        hamm84(buf+3+6*i, &err);
                        b3 = hamm84(buf+6+6*i, &err) * 16 +
                        hamm84(buf+5+6*i, &err);
                        if (err == 1)
                            return;
                        break;
                    default:
                        return;
                }

                int x = (b2 >> 7) | ((b3 >> 5) & 0x06);
                ttpage->floflink[i] = ((magazine ^ x) ?: 8) * 256 + b1;
                ttpage->flof = 1;
            }
            break;

        case 31: // private streams
            break;

        default: /// other packet codes...
            break;
    }
}

/** \fn SRTTypeTeletext::PageUpdated(int,int)
 *  \brief Updates the page, if given pagenumber is the same
 *         as the current shown page
 */
void SRTTypeTeletext::PageUpdated(int page, int subpage)
{
    if (!m_displaying)
        return;

    if (page != m_curpage)
        return;

    if (subpage != m_cursubpage && m_cursubpage != -1)
        return;

    if (m_output_device != NULL &&
        m_output_device->Is_Enabled())
    {
        const TeletextSubPage *ttpage = FindSubPage(m_curpage, m_cursubpage);

        if (ttpage)
        {
            QString new_line;
            bool is_double = false;
            for (int y = 1; y < 25; y++)
            {
                if (is_double == false)
                {
                    is_double = DecodeLine(ttpage->data[y], y, ttpage->lang, new_line);
                }
                else    
                {
                    is_double = false;
                }
            }
            m_output_device->Set_Line(new_line);
        }
    }        
}

/** \fn SRTTypeTeletext::HeaderUpdated(unsigned char*,int)
 *  \brief Updates the header (given in page with language lang)
 * 
 *  \param page Pointer to the header which should be displayed
 *  \param lang Language of the header
 *
 */
void SRTTypeTeletext::HeaderUpdated(unsigned char *page, int lang)
{
    (void)lang;

    if (!m_displaying)
        return;

    if (page == NULL)
        return;

    if (m_curpage_showheader == false)
        return;

    m_header_changed = true;
}

/** \fn SRTTypeTeletext::FindPageInternal(int, int)
 *  \brief Finds the given page
 *
 *  \param page Page number
 *  \param direction find page before or after the given page
 *  \return TeletextPage (NULL if not found) 
 */
const TeletextPage *SRTTypeTeletext::FindPageInternal(
    int page, int direction) const
{
    int mag = MAGAZINE(page);

    if (mag > 8 || mag < 1)
        return NULL;

    int_to_page_t::const_iterator pageIter;
    pageIter = m_magazines[mag - 1].pages.find(page);
    if (pageIter == m_magazines[mag - 1].pages.end())
        return NULL;

    const TeletextPage *res = &pageIter->second;
    if (direction == -1)
    {
        --pageIter;
        if (pageIter == m_magazines[mag - 1].pages.end())
        {
            int_to_page_t::const_reverse_iterator iter;
            iter = m_magazines[mag - 1].pages.rbegin();
            res = &iter->second;
        }
        else
            res = &pageIter->second;
    }

    if (direction == 1)
    {
        ++pageIter;
        if (pageIter == m_magazines[mag - 1].pages.end())
        {
            pageIter = m_magazines[mag - 1].pages.begin();
            res = &pageIter->second;
        }
        else
            res = &pageIter->second;
    }

    return res;
}

/** \fn SRTTypeTeletext::FindSubPageInternal(int, int, int) const
 *  \brief Finds the given page
 *
 *  \param page Page number
 *  \param subpage Subpage number (if set to -1, find the first subpage)
 *  \param direction find page before or after the given page
 *         (only if Subpage is not -1)
 *  \return TeletextSubPage (NULL if not found) 
 */
const TeletextSubPage *SRTTypeTeletext::FindSubPageInternal(
    int page, int subpage, int direction) const
{
    int mag = MAGAZINE(page);

    if (mag > 8 || mag < 1)
        return NULL;

    int_to_page_t::const_iterator pageIter;
    pageIter = m_magazines[mag - 1].pages.find(page);
    if (pageIter == m_magazines[mag - 1].pages.end())
        return NULL;

    const TeletextPage *ttpage = &(pageIter->second);
    int_to_subpage_t::const_iterator subpageIter =
        ttpage->subpages.begin();

    // try to find the subpage given, or first if subpage == -1
    if (subpage != -1)
        subpageIter = ttpage->subpages.find(subpage);

    if (subpageIter == ttpage->subpages.end())
        return NULL;

    if (subpage == -1)
        return &(subpageIter->second);

    const TeletextSubPage *res = &(subpageIter->second);
    if (direction == -1)
    {
        --subpageIter;
        if (subpageIter == ttpage->subpages.end())
        {
            int_to_subpage_t::const_reverse_iterator iter =
                ttpage->subpages.rbegin();
            res = &(iter->second);
        }
        else
        {
            res = &(subpageIter->second);
        }
    }

    if (direction == 1)
    {
        ++subpageIter;
        if (subpageIter == ttpage->subpages.end())
            subpageIter = ttpage->subpages.begin();

        res = &(subpageIter->second);
    }

    return res;
}

/** \fn SRTTypeTeletext::SetPage(uint)
 *  \brief Set TeletextPage to display
 *  \param page PageNr. (in hex)
 *  \param subpage Subpage (-1 if first available)
 */
void SRTTypeTeletext::SetPage(int page, int subpage)
{
    if (page < 0x100 || page > 0x899)
        return;

    m_pageinput[0] = (page / 256) + '0';
    m_pageinput[1] = ((page % 256) / 16) + '0';
    m_pageinput[2] = (page % 16) + '0';

    m_curpage = page;
    m_cursubpage = subpage;
    PageUpdated(m_curpage, m_cursubpage);
}

/** \fn cvt_char(char,int)
 *  \brief converts the given character to the given language
 */
static char cvt_char(char ch, int lang)
{
    int c = 0;
    for (int j = 0; j < 14; j++)
    {
        c = ch & 0x7F;
        if (c == lang_chars[0][j])
            ch = lang_chars[lang + 1][j];
    }
    return ch;
}

bool SRTTypeTeletext::DecodeLine(const unsigned char *page,
                                 uint row, int lang, QString &line) const
{
    bool mosaic;
    bool conceal;
    bool seperation;
    bool flash;
    bool doubleheight;
    bool blink;
    bool hold;
    bool endbox;
    bool startbox;

    char last_ch = ' ';
    char ch;
    QString row_text;

    bool isBlank = true;

  	for (uint x = (row == 1 ? 8 : 0);
         x < (uint)kTeletextColumns;
         ++x)
	{
        ch = page[x] & 0x7F;
        if (ch != ' ')
        {
            isBlank = false;
            break;
        }
	}
    if (isBlank)
    {
        return false;
    }
    mosaic = false;
    seperation = false;
    conceal = false;
    flash = false;
    doubleheight = false;
    blink = false;
    hold = false;
    endbox = false;
    startbox = false;
    uint flof_link_count = 0;

    for (uint x = (row == 1 ? 8 : 0); x < (uint)kTeletextColumns; ++x)
    {
        ch = page[x] & 0x7F;
        switch (ch)
        {
            case 0x00 ... 0x07: // alpha + foreground color
                mosaic = false;
                conceal = false;
                // increment FLOF/FastText count if menu item detected
                flof_link_count += (row == 25) ? 1 : 0;
                goto ctrl;
            case 0x08: // flash
                // XXX
                goto ctrl;
            case 0x09: // steady
                flash = false;
                goto ctrl;
            case 0x0a: // end box
                endbox = true;
                goto ctrl;
            case 0x0b: // start box
                goto ctrl;
            case 0x0c: // normal height
                doubleheight = false;
                goto ctrl;
            case 0x0d: // double height
                doubleheight = row < kTeletextRows-1;
                goto ctrl;
            case 0x10 ... 0x17: // graphics + foreground color
                mosaic = true;
                conceal = false;
                goto ctrl;
            case 0x18: // conceal display
                conceal = true;
                goto ctrl;
            case 0x19: // contiguous graphics
                seperation = false;
                goto ctrl;
            case 0x1a: // separate graphics
                seperation = true;
                goto ctrl;
            case 0x1c: // black background
                goto ctrl;
            case 0x1d: // new background
                goto ctrl;
            case 0x1e: // hold graphics
                hold = true;
                goto ctrl;
            case 0x1f: // release graphics
                hold = false;
                goto ctrl;
            case 0x0e: // SO (reserved, double width)
            case 0x0f: // SI (reserved, double size)
            case 0x1b: // ESC (reserved)
                ch = ' ';
                break;
            ctrl:
                ch = ' ';
                if (hold && mosaic)
                    ch = last_ch;
            break;

            case 0x80 ... 0x9f: // these aren't used
                ch = ' '; // BAD_CHAR;
                break;
            default:
                if (conceal && !m_revealHidden)
                    ch = ' ';
                break;
        }

        // Hide FastText/FLOF menu characters if not available
        if (flof_link_count && (flof_link_count <= 6))
        {
            const TeletextSubPage *ttpage =
                FindSubPage(m_curpage, m_cursubpage);

            if (ttpage)
            {
                bool has_flof = ttpage->floflink[flof_link_count - 1];
                ch = (has_flof) ? ch : ' ';
            }
        }
        if ((row != 0) || (x > 7))
        {
            if ((mosaic) && (ch < 0x40 || ch > 0x5F))
            {
//                DrawMosaic(surface, x, row, ch, doubleheight);
            }
            else
            {
                row_text += cvt_char(ch, lang);
            }
        }
    }
    if (row_text.length() > 0)
    {
        line += row_text.stripWhiteSpace();
        line += "\r\n";
    }
    return doubleheight;
}


SRTOutputFile::SRTOutputFile(QString outfile)
              :m_opened(false),
               m_enabled(true),
               m_outfp(outfile),
               m_line_dts(0),
               m_index(1)
{
}

SRTOutputFile::~SRTOutputFile(void)
{
    Close();
}

void
SRTOutputFile::Set_Enable(bool enabled)
{
    if (m_enabled != enabled)
    {
        m_enabled = enabled;
        if (enabled == false)
        {
            Output_Line();
        }
    }
}

void
SRTOutputFile::Set_Line(QString new_line)
{
    if (m_enabled)                                                      // make sure enabled for output
    {
        if (m_line.length() > 0)                                        // see if there is an existing subtitle
        {
            Output_Line();                                              // and if so output it now
        }
        m_line = new_line;                                              // save the new line
        m_line_dts = m_dts;                                             // and record it's starting dts
    }
}

QString
SRTOutputFile::Pts_To_Srt_Time(int64_t pts)
{
    bool is_neg = false;
    if (pts < 0)
    {
        pts = -pts;
        is_neg = true;
    }
    QString msg;
    return(msg.sprintf("%s%02d:%02d:%02d,%03d", (is_neg) ? "-" : "",
                (unsigned int)(pts / 90000.) / 3600,
                ((unsigned int)(pts / 90000.) % 3600) / 60,
                ((unsigned int)(pts / 90000.) % 3600) % 60,
                (((unsigned int)(pts / 90.) % 3600000) % 60000) % 1000));
}

void
SRTOutputFile::Output_Line()
{
    if (m_opened == true &&
        m_line.length() > 0)                                            // see if a line to output
    {
        m_outstream << QString("%1").arg(m_index) << "\r\n"
                    << Pts_To_Srt_Time(m_line_dts) << " --> " << Pts_To_Srt_Time(m_dts) << "\r\n" 
                    << m_line << "\r\n";
        m_index++;
        m_line = QString("");
    }
}

bool
SRTOutputFile::Open()
{
    if (m_opened == false)
    {
        if (!m_outfp.open(IO_WriteOnly))
        {
            VERBOSE(MPF_IMPORTANT, "Failed to open the subtitle SRT file");
            return false;
        }
        m_outstream.setDevice(&m_outfp);
        m_opened = true;
    }
    return true;
}

void
SRTOutputFile::Close()
{
    if (m_opened == true)
    {
        Output_Line();
        m_outfp.close();
        m_opened = false;
    }
}

